package de.umr

import de.umr.core.dataStructures.*
import de.umr.fixcon.*
import de.umr.fixcon.GlobalData.G
import de.umr.fixcon.GlobalData.f
import de.umr.fixcon.GlobalData.k
import de.umr.fixcon.GlobalData.secondsElapsed
import org.jgrapht.Graphs
import org.jgrapht.Graphs.neighborSetOf

/**This method contains the actual "algorithm".
 * After running the heuristic and setting up helper-objects for pruning-rules, the first while-loop finds different start-vertices
 * and builds a search-tree from each. The while-loop within builds the said search-tree.
 */
fun solve(timeLimit: Int = Int.MAX_VALUE): Pair<Solution<Int>, Double> {
    G.removeComponentsSmallerThreshold(k)           // Components of size less than k would be messy in heuristic

    var (sol, heuristicIsOptimal) = getStartingSolution()
    if (heuristicIsOptimal) return Pair(sol, secondsElapsed()) // program is already finished

    //Twins (only computed if used later)
    val (critPartition, openTwins) = if (f.allowsTwinRule) getCriticalPartitioning() else (Pair(null, null))
    if (f.allowsTwinRule) trimTwinSets(critPartition!!, openTwins!!)

    val domination_structure =
        if (f.edgeMonotone) domDAG(G, critPartition!!) else G.vertexSet().associateWith { HashSet() }

    /**MAIN-LOOP
     *
     * Each iteration of this loop builds one search-tree. This is done by picking a start-vertex and building it into
     * a subgraph of size [k] via a recursive algorithm, implemented iteratively with stacks.
     * After the search-tree has been explored, the start-vertex gets deleted in [G] so that no subgraph containing it
     * can ever be built again.
     */
    while (sol.value < f.globalOptimum() && G.vertexCount >= k) {

        // create all fields used in the next search-tree
        val startVertex = G.vertexSet().random()
        val subgraph = OrderedGraph<Int>().apply { addVertex(startVertex) } // also tracks order of added vertices

        val extension = SegmentedList<Int>().apply {
            this += Graphs.neighborListOf(G, startVertex)
        } // tracks neighbours first by a vertex for each vertex in current subgraph
        val pointers = mutableListOf(0) // list of indices storing which vertices form *extension* are added next
        val (twinBlacklist, dominationBlacklist, cliqueCompanion) = createSpeedupHelpers(startVertex)

        /**SEARCH-TREE-LOOP
         *
         * loops through nodes in the search-tree
         */
        while (pointers.isNotEmpty()) {

            // check if current subgraph should be reduced by 1 vertex
            if (checkIfBacktrack(pointers, extension, subgraph, sol, cliqueCompanion)) {
                if (numVerticesMissing(subgraph) > 0) extension.removeLastSegment()
                val poppedVertex = subgraph.removeLastVertex()

                if (f.allowsTwinRule) {
                    twinBlacklist!!.removeLastSegment()
                    twinBlacklist.addToLast(critPartition!![poppedVertex])
                }

                if (f.edgeMonotone) {
                    dominationBlacklist!!.removeLastSegment()
                    dominationBlacklist.addToLast(domination_structure[poppedVertex]!!)
                }

                cliqueCompanion.removeVertex(poppedVertex)

                pointers.removeLast()

            } else if (f.edgeMonotone && extension[pointers.last()] in dominationBlacklist!!) {
                pointers[pointers.size - 1]++ //move top pointer 1 forward
                AlgoStats.dominationDagSkips++

            } else if (f.allowsTwinRule && extension[pointers.last()] in twinBlacklist!!) {
                pointers[pointers.size - 1]++ //move top pointer 1 forward
                AlgoStats.criticalTwinSkips++

            } else {  // grow search-tree by 1 vertex
                if (secondsElapsed() >= timeLimit) {
                    println("PROGRAM RAN OUT OF TIME !!!!!")
                    return Pair(sol, secondsElapsed())
                }
                printNodeCount()

                val nextVertex = extension[pointers.last()]
                if (numVerticesMissing(subgraph) > 1) extension += Graphs.neighborListOf(G, nextVertex)
                    .filter { it !in extension && it != startVertex }

                if (f.allowsTwinRule)
                    twinBlacklist!! += emptyList()
                if (f.edgeMonotone)
                    dominationBlacklist!! += emptyList()

                subgraph.expandSubgraph(nextVertex)

                cliqueCompanion.expandSubgraph(nextVertex)

                pointers[pointers.size - 1]++ // increment last pointer
                pointers.add(pointers.last())
            }

            if (numVerticesMissing(subgraph) == 0 && f.eval(subgraph) > sol.value) //update if a better solution is found
                sol = Solution(subgraph.copy(), f.eval(subgraph))
        }
        val start_vertex_neighbors = neighborSetOf(G, startVertex) subtract (critPartition!![startVertex])
        deleteSearchTreeRoot(critPartition, startVertex)
        merge_domDAG(G, critPartition, domination_structure, start_vertex_neighbors)
    } // end main-loop
    return Pair(sol, secondsElapsed())
} //end solve-method

private fun checkIfBacktrack(
    pointers: MutableList<Int>,
    extension: SegmentedList<Int>,
    subgraph: OrderedGraph<Int>,
    sol: Solution<Int>,
    cliqueCompanion: CJR_Tracker
) =
    pointers.last() >= extension.size ||
            numVerticesMissing(subgraph) == 0 ||
            (vertexAdditionRule(subgraph, sol)) ||
            cliqueJoinRule(pointers, extension, subgraph, cliqueCompanion, sol) ||
            (f.edgeMonotone && universalGraphRule(subgraph, sol.value, extendable(pointers, extension, subgraph)))

private fun createSpeedupHelpers(startVertex: Int): Triple<SegmentedList<Int>?, SegmentedList<Int>?, CJR_Tracker> {
    val twinBlacklist = if (f.allowsTwinRule) SegmentedList<Int>().apply { repeat(2) { this += HashSet() } } else null
    val dominationBlacklist =
        if (f.edgeMonotone) SegmentedList<Int>().apply { repeat(2) { this += HashSet() } } else null
    val cliqueCompanion = CJR_Tracker(k = k).apply { addVertex(startVertex) }
    return Triple(twinBlacklist, dominationBlacklist, cliqueCompanion)
}

private fun printNodeCount() {
    if (++AlgoStats.searchTreeNodes % 1_000_000 == 0L)
        println("tree-nodes:".padEnd(Settings.paddingRight) + "${AlgoStats.searchTreeNodes / 1_000_000} million")
}

private fun deleteSearchTreeRoot(critPartition: Partitioning<Int>?, startVertex: Int) {
    if (f.allowsTwinRule)
        deleteUpdateTwinSet(critPartition!!, startVertex)
    else // crit-partition isn`t even computed
        G.removeVertex(startVertex)
}

private fun getStartingSolution(): Pair<Solution<Int>, Boolean> {
    val sol = if (Settings.useHeuristic) getHeuristic() else Solution()

    val optimal = if (sol.value == f.globalOptimum()) {
        println("Heuristic was optimal")
        true
    } else {
        println("Heuristic not optimal")
        false
    }

    return Pair(sol, optimal)
}


private fun extendable(pointers: MutableList<Int>, extension: SegmentedList<Int>, subgraph: OrderedGraph<Int>) =
    ArrayList<Int>().apply {
        for (i in pointers.indices.reversed())
            if (extension.segmentSizes[i] > pointers.last()) add(subgraph.orderedVertices[i])
            else break
    }.filter { subgraph.degreeOf(it) < G.degreeOf(it) }


/**This function is difficult to move into file "PruningRules" as the Clique-Join-Rule depends on *cliqueCompanion* and
 * *extendable()* which is in the local scope of the main-loop.
 */
fun cliqueJoinRule(
    pointers: MutableList<Int>,
    extension: SegmentedList<Int>,
    subgraph: OrderedGraph<Int>,
    cliqueCompanion: CJR_Tracker,
    sol: Solution<Int>
): Boolean {
    if (f.isEdgeWeighted) {
        return false
        //TODO anpassen
    } else {
        if (!f.edgeMonotone) return false
        cliqueCompanion.connectVertices(extendable(pointers, extension, subgraph), cliqueCompanion.cliqueVertices())
        val isApplicable = f.eval(cliqueCompanion) <= sol.value
        cliqueCompanion.disconnectVertices(extendable(pointers, extension, subgraph), cliqueCompanion.cliqueVertices())
        return isApplicable.also { if (it) AlgoStats.cliqueJoinRuleSkips++ }
    }
}

fun numVerticesMissing(subgraph: OrderedGraph<Int>) = k - subgraph.vertexCount