package de.umr.core

import de.umr.core.dataStructures.degrees
import de.umr.core.dataStructures.edgeCount
import de.umr.core.dataStructures.vertexCount
import de.umr.fixcon.Settings
import org.jgrapht.Graph
import org.jgrapht.graph.DefaultEdge

/**
 * Provides functions to analyze graphs to provide user feedback and for debugging.
 */


/**The keys in the map represent the degree, the values are the share of this degree in [g].
 *
 * If 60% of vertices in the graph have degree 2 and 40% have degree 3, the map will be like {2: 0.6, 3: 0.4}*/
fun <V> degreeShareMap(g: Graph<V, DefaultEdge>): Map<Int, Double> = g.vertexSet()
        .groupingBy { g.degreeOf(it) }.eachCount()
        .mapValues { 1.0 * it.value / g.vertexCount }

/**Prints a summary of [g], including number of vertices and edges, maximum degree and the overall shares of the
 * different vertex-degrees that are present.*/
fun <V> printAnalysis(g: Graph<V, DefaultEdge>, tableSize: Int = 3) {
    println("Vertices:".padEnd(Settings.paddingRight) + g.vertexCount)
    println("Edges:".padEnd(Settings.paddingRight) + g.edgeCount)
    println("Max-Degree:".padEnd(Settings.paddingRight) + g.degrees.maxOrNull()!!)
    degreeShareMap(g).entries.sortedByDescending { it.value }.take(tableSize).forEach {
        println("${it.key}".padEnd(Settings.paddingRight) + ("%.1f".format(it.value * 100) + " % ").padStart(0))
    }
}
