package de.umr.core

import de.umr.fixcon.Settings
import org.jgrapht.graph.DefaultEdge
import org.jgrapht.graph.SimpleWeightedGraph
import java.io.File
import java.lang.Integer.parseInt

/**This file provides Input-Output related services.
 * It stores paths of files containing commonly used graphs, and has methods to read in files representing graphs.
 *
 * Files are stored in an edge-list format: If a line says "13 47", this means that there are two vertices with respective
 * names "13" and "47", and those two vertices are connected by an undirected edge.
 * The line "13 47 2.5" would mean  that the edge has weight 2.5.
 **/


/**Network repository is a website which provides files for graphs. Files downloaded from there
 * are at this folder.*/
const val repo = "../data/network repository/"

const val konect = "../data/konect/"

/**Stores some constants for paths in this project as strings.
 *
 * The class [GraphFile] contains paths to graphs commonly used in the tests, so that their path
 * doesn't has to be written out redundantly.
 *
 * Some files are stored in this repository in the package *data*.
 * A bigger selection of graphs is in the data-repository of the algorithms research group.
 * These files can be accessed under the paths [repo] and [konect].
 *
 */
enum class GraphFile(val path: String) {
    CustomGraph1("./data/customGraph1.txt"),
    InfUsAir("./data/inf-USAir97.mtx"),
    BioDmela("${repo}bio/bio-dmela.mtx"),
    InfPower("${repo}infrastructure/inf-power.mtx"),
    SocBrightkite("${repo}social/soc-brightkite/soc-brightkite.mtx"),
    OutDolphins("${konect}undirected-simple-small/dolphins/out.dolphins"),
    InfEuroRoad("${repo}infrastructure/inf-euroroad.edges"),
    Sample("./data/sample"),
    CaSandiAuths("${repo}collaboration/ca-sandi_auths.mtx"),
    Hamming10_4("./data/hamming10-4.mtx"),
    PHat_1500_3("./data/p-hat1500-3.mtx"),
    InfOpenFlights("./data/inf-openflights.edges"),
    CoPapersCiteseer("./data/coPapersCiteseer.mtx"),
    CustomTree("./data/CustomTree.txt"),
    BadGraph("./data/badGraph.txt"),
    SocAdvogato("${repo}social/soc-advogato/soc-advogato.txt"),
    MouseRetina("${repo}brain/bn-mouse_retina_1.edges"),
    Heart2("${repo}misc/heart2.edges"),
    BioYeast("${repo}bio/bio-yeast.mtx"),
    MorenoZebra("${konect}undirected-simple-small/moreno_zebra/out.moreno_zebra_zebra"),
    UcidataZachary("${konect}undirected-simple-small/ucidata-zachary/out.ucidata-zachary"),
    Celegans("$repo/bio/bio-celegans.mtx"),
    Diseasome("$repo/bio/bio-diseasome.mtx"),
    MixedSpecies("$repo/brain/bn-cat-mixed-species_brain_1.edges"),
    CSphd("$repo/collaboration/ca-CSphd.mtx")
}

/**Line is not commented out (doesn't start with % or #) and is not empty.*/
val validLine: (String) -> Boolean = { !it.startsWith("%") && !it.startsWith("#") && it.isNotEmpty() }

/**Reads in the graph at [file] and returns a list of edges, where each edge is in the format <VertexID, VertexID, Weight>.*/
fun edgesFromFile(file: String, allowLoops: Boolean = false): List<Triple<Int, Int, Double>> = File(file).readLines()
        .filter { validLine(it) }
        .map { it.split(Regex("""\s+""")) } //split vertex-IDs by whitespace between them
        .filter { it[0] != it[1] || allowLoops } //no loops if not allowed (like we ever would ever do so ...)
        .mapTo(ArrayList()) { Triple(parseInt(it[0]), parseInt(it[1]), Settings.defaultEdgeWeight) }

fun edgesFromFile(file: GraphFile, allowLoops: Boolean = false) = edgesFromFile(file.path, allowLoops)

fun graphFromFile(file: String, allowLoops: Boolean = false) = graphFromWeightedEdges(edgesFromFile(file, allowLoops))

fun graphFromFile(f: GraphFile): SimpleWeightedGraph<Int, DefaultEdge> = graphFromWeightedEdges(edgesFromFile(f))
