package de.umr.core.dataStructures

import de.umr.core.graphFromWeightedEdges
import de.umr.core.intersectAll
import de.umr.fixcon.GlobalData.G
import de.umr.fixcon.vertexHash
import org.jgrapht.Graph
import org.jgrapht.Graphs.*
import org.jgrapht.alg.connectivity.ConnectivityInspector
import org.jgrapht.graph.DefaultEdge

/**
 * Adds new methods to the *Graph*-class from jGraphT.
 */


/**Returns the number of all vertices (even isolated ones).*/
val <V> Graph<V, DefaultEdge>.vertexCount: Int get() = vertexSet().size

/**Returns the number of all edges in the graph.*/
val <V> Graph<V, DefaultEdge>.edgeCount get() = edgeSet().size

/**Returns a list of all vertex-degrees in the graph. This list is unordered, but if degree *x* appears *n* times
 * in the graph, then *x* also appears *n* times in the list.*/
val <V> Graph<V, DefaultEdge>.degrees get() = vertexSet().map { degreeOf(it) }

/**Returns the neighbour-vertices of [vertices]. This is defined as the set of all vertices *v* so that *v* has an edge
 * to some vertex in [vertices], but *v* is not contained in [vertices].*/
fun <V> Graph<V, DefaultEdge>.neighbours(vertices: Collection<V>): Set<V> =
        vertices.flatMapTo(HashSet()) { neighborSetOf(this, it) }.apply { removeAll(vertices) }

/**The closed neighbourhood of a vertex *v* is the set of *v* and all vertices with an edge to *v*.*/
fun <V> Graph<V, DefaultEdge>.closedNB(v: V): Set<V> = neighborSetOf(this, v).apply { add(v) }

/**Checks fast if [v1] and [v2] have the same open neighbourhood.
 * Method instantly returns false if degrees aren't equal, as this is a necessary condition.*/
fun <V> Graph<V, DefaultEdge>.openNBEqualsFast(v1: V, v2: V) =
        degreeOf(v1) == degreeOf(v2) && !containsEdge(v1, v2) && neighborSetOf(this, v1) == neighborSetOf(this, v2)

/**Checks fast if [v1] and [v2] have the same closed neighbourhood.
 * Method instantly returns false if degrees aren't equal, as this is a necessary condition.*/
fun <V> Graph<V, DefaultEdge>.closedNBEqualsFast(v1: V, v2: V) =
        degreeOf(v1) == degreeOf(v2) && containsEdge(v1, v2) && closedNB(v1) == closedNB(v2)

/**Adds an edge between [v1] and [v2] with weight [weight]
 * @return The resulting [Graph] after the call of this method*/
fun <V> Graph<V, DefaultEdge>.addWeightedEdge(v1: V, v2: V, weight: Double): Graph<V, DefaultEdge> =
        also { addEdgeWithVertices(this, v1, v2);setEdgeWeight(v1, v2, weight) }

/**This method was created because from my knowledge jGraphT doesn't provide copying of graphs.
 * It's only guaranteed to work with primitives as vertices, because if the method was totally generic it could not be
 * guaranteed that the vertices are cloneable.
 * @return A new Integer-valued graph, which is a copy of [this] graph.*/
fun <V> Graph<V, DefaultEdge>.copy() = graphFromWeightedEdges(edgeSet().map { getEdgeSource(it) to getEdgeTarget(it) }
        .map { (s, t) -> Triple(s, t, getEdgeWeight(getEdge(s, t))) })

/**Logically assumes [this] graph is a subgraph of [G].
 *
 * [newVertex] is added to [this] a graph and an edge is added between [newVertex] and each vertex in [this] if
 * there is also an edge between them in [G].*/
fun Graph<Int, DefaultEdge>.expandSubgraph(newVertex: Int) =
        intersectAll(listOf(neighborSetOf(G, newVertex), vertexSet())).forEach { addEdgeWithVertices(this, newVertex, it) }

/**Note: jGraphT also provides this method, but it counts ALL triangles and doesn't stop once it has found one.
 *
 * @return **True** iff the graph contains at least 1 triangle (a clique with 3 vertices).*/
fun <V> Graph<V, DefaultEdge>.hasTriangle() =
        edgeSet().any { (neighborSetOf(this, getEdgeSource(it)) intersect neighborSetOf(this, getEdgeTarget(it))).isNotEmpty() }

/**Removes all maximally connected components of the graph that have fewer than [threshold] vertices.*/
fun <V> Graph<V, DefaultEdge>.removeComponentsSmallerThreshold(threshold: Int) =
        ConnectivityInspector(this).connectedSets().filter { it.size < threshold }.forEach { removeAllVertices(it) }

/**Adds an edge from each vertex of [vCol1] to each vertex of [vCol2].*/
fun <V> Graph<V, DefaultEdge>.connectVertices(vCol1: Collection<V>, vCol2: Collection<V>) {
    for (v1 in vCol1) for (v2 in vCol2) addEdge(v1, v2)
}

/**Removes the edge from each vertex of [vCol1] to each vertex of [vCol2] in case it exists.*/
fun <V> Graph<V, DefaultEdge>.disconnectVertices(vCol1: Collection<V>, vCol2: Collection<V>) {
    for (v1 in vCol1) for (v2 in vCol2) removeEdge(v1, v2)
}

/**Displays the graph in a human-readable format. The neighbors of each vertex are listed.
 * The path 0-1-2 would result in "0: [1]\n1: [0, 2]\n2: [1]".*/
fun <V> Graph<V, DefaultEdge>.toPrettyString(): String {
    return vertexSet().joinToString("\n") { "$it: ${neighborListOf(this, it)}" }
}

/**
 * @return A list of the vertices in *g*, sorted in descending order by their [vertexHash].
 */
fun Graph<Int, DefaultEdge>.verticesSortedByHash(): List<Int> {
    val vertexHashes: Map<Int, List<Int>> = vertexSet().associateWith { vertexHash(this, it) { v -> neighborSetOf(this, v) } } // caches the hash-value for each vertex

    // First compare by degree, then resolve draws by sum of degrees of neighbours, then resolve draws by sum of hashes of neighbours
    return vertexSet().sortedWith(compareBy({ vertexHashes[it]!![0] }, { vertexHashes[it]!![1] }, { vertexHashes[it]!![2] })).reversed()
}