package de.umr.core.dataStructures

import de.umr.core.fromVertices
import de.umr.fixcon.Settings
import de.umr.fixcon.Settings.defaultEdgeWeight
import org.jgrapht.graph.DefaultEdge
import org.jgrapht.graph.SimpleWeightedGraph

/**This class extends [SimpleWeightedGraph], but it additionally tracks the order in which the vertices were added.
 *
 * You can only remove that vertex that was added last using [removeLastVertex], like in a stack.
 * [removeVertex] is depreciated now (will throw Exception), use [removeLastVertex] instead.*/
class OrderedGraph<V>(original: SimpleWeightedGraph<V, DefaultEdge> = fromVertices()) : SimpleWeightedGraph<V, DefaultEdge>(DefaultEdge::class.java) {

    val orderedVertices = ArrayList<V>()        // tracks the order of insertion of the vertices.

    /**Works as flag for [removeVertex],so that it can only used by [removeLastVertex] and not from the user (will throw Exception).
     *
     * To use [removeVertex] inside of [removeLastVertex], [internalChangeFlag] is set to *true*, then [removeVertex]
     * is performed and [internalChangeFlag] is immediately set to *false* again.*/
    private var internalChangeFlag: Boolean = false

    init {      // copies the graph provided as constructor-argument
        original.edgeSet().forEach { addWeightedEdge(original.getEdgeSource(it), original.getEdgeTarget(it), defaultEdgeWeight) }
    }

    /**Nothing changes except that it's tracked that [v] is the vertex added most recently.
     * This way it can be removed with [removeLastVertex].
     *
     * @return *True* if the graph changed as result (<=> [v] is a new vertex).*/
    override fun addVertex(v: V) = super.addVertex(v).also { if (it) orderedVertices.add(v) }

    /**Removes the vertex that was added most recently to the graph.
     * @return The vertex that was removed.*/
    fun removeLastVertex(): V {
        require(orderedVertices.isNotEmpty()) // graph is not empty
        val v = orderedVertices.removeLast()
        internalChangeFlag = true // makes call of *removeVertex* possible
        removeVertex(v)
        return v
    }

    /**
     * Removal of arbitrary vertices is not supported. This method is depreciated and will throw exception.
     * @See [removeLastVertex].*/
    override fun removeVertex(v: V): Boolean {
        require(internalChangeFlag) { "This method may not be called from the outside." }
        internalChangeFlag = false
        return super.removeVertex(v)
    }
}