package de.umr.fixcon

import org.jgrapht.graph.DefaultEdge
import org.jgrapht.graph.SimpleWeightedGraph
import org.paukov.combinatorics3.Generator.combination

/**
 * Object that helps for testing the Clique-Join-Rule (CJR).
 * The CJR is only applicable for edge-monotone functions. Let *j* be *k* - (current number of vertices).
 * CJR checks the value of the objective function if all *j* vertices would be a clique and all of them would also be
 * connected to all extendable vertices in the subgraph. This value then has to be the best possible value the current
 * subgraph could ever attain. Therefore, if this value is not better than the current best solution, we can backtrack.
 *
 * The CJR_Tracker therefore "shadows" the actual subgraph of the algorithm, except that it always also has a clique of
 * size *j*. For this to work, you need to update [CJR_Tracker] with [addVertex] and [removeVertex] just like you do
 * the real subgraph.
 *
 * One problem when creating this class was that new IDs for the vertices in the clique are needed.
 * Currently, the *j* vertices have IDs -1, ..., -*j*. This assumes that the real IDs you use in [addVertex] are all
 * non-negative (which is also checked).
 *
 * @param k The size of the subgraph in the FixCon-problem.
 */
class CJR_Tracker(val k: Int) : SimpleWeightedGraph<Int, DefaultEdge>(DefaultEdge::class.java) {

    var currentCliqueSize = k.also { require(it >= 0) }

    fun cliqueVertices() = (-1 downTo -currentCliqueSize).toList() //IDs for the current vertices in the clique

    init {
        cliqueVertices().forEach { super.addVertex(it) }
        combination(cliqueVertices()).simple(2).forEach { super.addEdge(it[0], it[1]) } //connected all pairwise
    }

    /**
     * @param v The ID of the new vertex
     * @return True iff [v] was not already present (like specified in superclass).
     */
    override fun addVertex(v: Int): Boolean {
        require(v >= 0) { "ID must be non-negative" }
        require(currentCliqueSize >= 0) { "Clique is already empty." }

        if (containsVertex(v)) return false
        super.removeVertex(-(currentCliqueSize--)) //clique shrinks by one vertex
        super.addVertex(v)
        return true
    }

    /**
     * @param v The ID of the new vertex
     * @return True iff [v] was not already present (like specified in superclass).
     */
    override fun removeVertex(v: Int): Boolean {
        require(v >= 0) { "ID must be non-negative" }
        if (!containsVertex(v)) return false

        super.addVertex(-(++currentCliqueSize)) //clique expands by one vertex
        (-1 downTo -(currentCliqueSize - 1)).forEach { addEdge(-currentCliqueSize, it) } //connect new clique-vertex to all others
        super.removeVertex(v)
        return true
    }
}