package de.umr.fixcon

import de.umr.core.dataStructures.*
import de.umr.fixcon.GlobalData.G
import de.umr.fixcon.GlobalData.k
import org.jgrapht.Graph
import org.jgrapht.Graphs.neighborListOf
import org.jgrapht.Graphs.neighborSetOf
import org.jgrapht.graph.DefaultEdge

/**
 * Maps each vertex to a hash (which is a list of length 3). Normally, the hash is an integer, so the entries of
 * the list would be added. In this program, the entries of the list are used in the domination-DAG,
 * so they are left unchanged.
 *
 * @return [degree, sum of degrees of neighbours, sum of hash-codes of neighbours]
 */
fun <V> vertexHash(graph: Graph<V, DefaultEdge>, v: V, nbSelector: (V) -> Set<V>): List<Int> =
        listOf(graph.degreeOf(v), nbSelector(v).sumOf { graph.degreeOf(it) }, nbSelector(v).sumOf { it.hashCode() })


fun getCriticalPartitioning(): Pair<Partitioning<Int>, Set<Int>> {

    fun partitionWithHash(vertices: Collection<Int>, hashFu: (Int) -> List<Int>, nbSelector: (Int) -> Collection<Int>): Partitioning<Int> =
            Partitioning<Int>().apply { vertices.groupBy { hashFu(it) }.values.forEach { addByEQPredicate(it) { x, y -> nbSelector(x) == nbSelector(y) } } }


    val partitioning = partitionWithHash(G.vertexSet(), { vertexHash(G, it) { v -> G.closedNB(v) } }, { G.closedNB(it) })

    val remainingVertices = partitioning.elements().filter { partitioning[it].size == 1 }.toSet()

    partitioning -= remainingVertices
    partitioning.disjointUnion(partitionWithHash(remainingVertices, { vertexHash(G, it) { v -> neighborSetOf(G, v) } }, { neighborSetOf(G, it) }))

    println("Biggest crit-set: ${partitioning.subsets().maxOf { it.size }}")
    return Pair(partitioning , remainingVertices)
}

fun  critCliqueMerge(partitioning: Partitioning<Int>, vertices: Collection<Int>) {
    for (v in vertices)
        for (nb in neighborListOf(G, v))
            if (partitioning[v] !== partitioning[nb] && G.closedNBEqualsFast(v, nb))
                partitioning.merge(v, nb)
}

fun  critISMerge(partitioning: Partitioning<Int>, vertices: Collection<Int>) {
    for (v1 in vertices) {
        val middleVertex: Int? = neighborListOf(G, v1).minByOrNull { G.degreeOf(it) }
        if (middleVertex != null)
            for (v2 in neighborListOf(G, middleVertex))
                if (partitioning[v1] !== partitioning[v2] && G.openNBEqualsFast(v1, v2))
                    partitioning.merge(v1, v2)
    }
}

fun deleteUpdateTwinSet(criticalPartition: Partitioning<Int>, startVertex: Int) {
    val nbVertices = G.neighbours(criticalPartition[startVertex])
    G.removeAllVertices(criticalPartition[startVertex])
    criticalPartition.removeSubset(startVertex)
    critCliqueMerge( criticalPartition, nbVertices)
    critISMerge( criticalPartition, nbVertices)
}

/**
 * This function is only called once at the start of the program. For each twin-set, it removes the vertices
 * that can already be detected as not needed by the size of the desired subgraph.
 *
 * Each solution containing exactly [k] vertices can contain at most [k] vertices of a closed (=true) twin set and
 * at most [k]−1 vertices of a open (=false) twin set.
 */
fun trimTwinSets(p: Partitioning<Int>, openTwins: Set<Int>) {
    var counter = 0
    for (sub in p.subsets()) {
        val reps = (sub.size - k) + if (sub.first() in openTwins) 1 else 0
        repeat(reps) {
            val v = sub.first()
            sub.remove(v)
            G.removeVertex(v)
            counter++
        }
    }
    println("twin-vertices pruned:".padEnd(Settings.paddingRight) + counter)
}