package de.umr.fixcon

import de.umr.core.dataStructures.Partitioning
import de.umr.core.dataStructures.closedNB
import de.umr.core.dataStructures.verticesSortedByHash
import org.jgrapht.Graph
import org.jgrapht.Graphs.neighborListOf
import org.jgrapht.Graphs.neighborSetOf
import org.jgrapht.graph.DefaultEdge

/**
 * Vertex **a** "dominates" vertex **b** if the open neighbourhood of **b** is a subset
 * of the closed neighbourhood of **a**. In this case the key **a** maps to a set which contains **b**.
 */
fun bruteForceDomDAG(g: Graph<Int, DefaultEdge>, critPartitioning: Partitioning<Int>): Map<Int, Set<Int>> {
    val map: Map<Int, MutableSet<Int>> =
        g.vertexSet().associateWith { HashSet() }  // each vertex has empty domination-set

    for (a in g.vertexSet()) {
        val closedNB_of_a = g.closedNB(a)
        for (b in g.vertexSet())
            if (b !in critPartitioning[a] && closedNB_of_a.containsAll(neighborListOf(g, b))) // test if it is a subset
                map[a]!!.add(b)
    }
    return map
}

fun domDAG(g: Graph<Int, DefaultEdge>, critPartitioning: Partitioning<Int>): Map<Int, MutableSet<Int>> {

    val roots: MutableSet<Int> = HashSet()
    val children: Map<Int, MutableSet<Int>> = g.vertexSet().associateWith { HashSet() }
    val domination_structure: Map<Int, MutableSet<Int>> = g.vertexSet().associateWith { HashSet() }

    val sorted_vertices = g.verticesSortedByHash()

    val visited_twins: MutableSet<Int> = HashSet()

    for (v in sorted_vertices) if (v !in visited_twins) {
        val x: Int = neighborListOf(g, v).minByOrNull { g.degreeOf(it) }!!
        var root_candidates: Set<Int> = (g.closedNB(x).toMutableSet().apply { remove(v) }) intersect roots

        for (u in neighborListOf(g, v)) {
            root_candidates = root_candidates intersect g.closedNB(u)
            if (root_candidates.isEmpty())
                break
        }

        if (root_candidates.isEmpty()) {
            roots.add(v)
        } else {

            val visited: MutableSet<Int> = HashSet()
            val Q = ArrayDeque(root_candidates)
            while (Q.isNotEmpty()) {
                val u = Q.removeFirst()
                for (w in children[u]!!)
                    if (g.degreeOf(v) < g.degreeOf(w))
                        if (g.closedNB(w).containsAll(neighborSetOf(g, v)))
                            if (w !in visited) {
                                Q.add(w)
                                visited.add(w)
                            }
                children[u]!!.add(v)
                for (a in critPartitioning[u])
                    for (b in critPartitioning[v])
                        domination_structure[a]!!.add(b)
            }
        }
        visited_twins.addAll(critPartitioning[v])
    }
    return domination_structure
}

/**vertex v gets deleted*/
fun merge_domDAG(
    g: Graph<Int, DefaultEdge>,
    critPartitioning: Partitioning<Int>,
    domination_structure: Map<Int, MutableSet<Int>>,
    neighbors_of_v: Set<Int>
) {
    val touched = HashSet<Int>()
    for (u in neighbors_of_v) {
        if (g.degreeOf(u) >= 1) {
            val cand = HashSet<Int>() // candidates of being dominators of u
            val min_u: Int = neighborSetOf(g, u).minByOrNull { g.degreeOf(it) }!!
            for (w in neighborSetOf(g, u)) {
                if (g.degreeOf(w) >= g.degreeOf(u) && w !in neighbors_of_v) {
                    cand.add(w)
                }
            }
            for (x in neighborSetOf(g, min_u)) {
                if (x != u && g.degreeOf(x) >= g.degreeOf(u) && x !in neighbors_of_v && x !in g.closedNB(u)) {
                    cand.add(x)
                }
            }

            for (c in cand) {
                var c_is_dominator = false
                if (u !in domination_structure[c]!!) {
                    if (c in neighborSetOf(g, u) && g.closedNB(c).containsAll(g.closedNB(u)))
                        c_is_dominator = true
                    if (c !in neighborSetOf(g, u) && g.closedNB(c).containsAll(neighborListOf(g, u)))
                        c_is_dominator = true
                }

                if (c_is_dominator) {
                    println("Dom-DAG merge executed")

                    domination_structure[c]!!.addAll(critPartitioning[u])
                    domination_structure[c]!!.addAll(domination_structure[u]!!)
                }
            }
            touched.addAll(critPartitioning[u])
        }
    }
}