package de.umr.fixcon

import de.umr.core.dataStructures.edgeCount
import de.umr.core.dataStructures.vertexCount
import de.umr.core.graphFromFile
import de.umr.fixcon.objectiveFunctions.AbstractGraphFunction
import de.umr.fixcon.objectiveFunctions.graphFunctionByID
import org.jgrapht.Graph
import org.jgrapht.graph.DefaultEdge
import kotlin.properties.Delegates

/**
 * Data-wrapper for all information specified by the user via the command line.
 *
 * args[0] == File-Path for the graph *G*.
 * args[1] == k
 * args[2] == functionID, param1, param2, ..., paramN-1, paramN
 * args[3] == time limit in seconds ()
 *
 * There are 2 main architectural approaches to handle scoping for an algorithmic solver:
 * - Try to enforce information-hiding like usual, resulting in redundant parameters for most methods
 * (nearly everything needs *G* or the current subgraph).
 * - Keep these things in global scope so that it's all shorter.
 *
 * I've chosen the latter for now. This is where this global state is stored.
 */
object GlobalData {

    /**
     * This method sets all the fields in this singleton.
     * Takes the String-Array "args" that's usually provided when the program is called via the console.
     */
    fun init(args: Array<String>) {
        /**First the stuff that is directly specified*/
        path = args[0]
        k = args[1].toInt()
        funcID = args[2].split(",").first().toInt()
        funcParams = args[2].split(",").drop(1).map { it.toInt() }
        timeout = args[3].toInt()

        /**Then the rest...*/
        G = graphFromFile(path).also { g -> require(g.vertexSet().all { v -> v >= 0 }) }
        f = graphFunctionByID(funcID, funcParams)

        vertexCount = G.vertexCount
        edgeCount = G.edgeCount
    }

    // All the fields that can be set with [initFromCommandLineArgs].
    lateinit var path: String                // path from where the input-graph is read
    lateinit var G: Graph<Int, DefaultEdge>  // input-graph, which is specified via a path to a file
    var funcID by Delegates.notNull<Int>()   // ID that specifies the objective function
    lateinit var f: AbstractGraphFunction    // objective function, specified via ID
    lateinit var funcParams: List<Int>       // parameters for the objective function (empty when not needed)
    var k by Delegates.notNull<Int>()        // size of the desired subgraph

    var vertexCount by Delegates.notNull<Int>() // the number of vertices the input-graph initially has. May be incorrect later during the algorithm as vertices get deleted.
    var edgeCount by Delegates.notNull<Int>()   // the same but with edges ...
    var timeout by Delegates.notNull<Int>()     // Maximal time the algorithm is allowed to run, specified in seconds from console

    private val startTime: Long = System.currentTimeMillis()  //Time tracking

    /**@return How many seconds the program is already running*/
    fun secondsElapsed(): Double = (System.currentTimeMillis() - startTime) / 1000.0 // Divides by 1000 to get seconds.
}