package de.umr.fixcon

import de.umr.core.dataStructures.*
import de.umr.core.fromVertices
import de.umr.core.intersectAll
import de.umr.fixcon.GlobalData.f
import de.umr.fixcon.GlobalData.G
import de.umr.fixcon.GlobalData.k
import org.apache.commons.math3.analysis.function.Inverse
import org.apache.commons.math3.distribution.EnumeratedDistribution
import org.apache.commons.math3.util.Pair
import org.jgrapht.Graphs.neighborListOf
import org.jgrapht.alg.connectivity.ConnectivityInspector
import kotlin.math.log2

/**Provides one random sample from a finite set of elements, which are specified as keys in [weightMap].
 * The elements may have different probabilities, as specified in the values of [weightMap].
 * The values in [weightMap] don't need to sum to one, they are normalized internally.
 *
 * If you want to transform the probability-weight, you can do so using [weightFunc]. E.g. if [weightFunc] is x -> x^2
 * then the [weightMap] {"a" : 2, "b" : 3} is treated like {"a" : 4, "b" : 9}.
 */
fun <T> takeRandom(weightMap: Map<T, Int>, weightFunc: (Int) -> Double = { 1.0 * it }): T =
        EnumeratedDistribution(weightMap.map { Pair(it.key, weightFunc(it.value)) }).sample()

/**Handles the entire heuristic. Performs multiple runs of local-search.
 * Uses different randomness-weights to fit both sparse and dense graphs.
 *
 * @return Some solution that may not be optimal, but is often pretty ok.
 */
fun getHeuristic(): Solution<Int> {

    val runs: Int = (5 + log2(G.vertexCount.toDouble()) * k).toInt()
    println("Number of heuristic runs: $runs")
    val sol = Solution<Int>()

    /**Heuristically generates a solution that can then be improved further with local-search.
     *
     * It builds the subgraph incrementally by starting just with [startVertex] and then repeatedly adding a vertex
     * with [extensionPicker].*/
    fun generateStartingSolution(startVertex: Int, extensionPicker: (MutableMap<Int, Int>) -> Int): Solution<Int> {
        val subgraph = fromVertices(startVertex)
        val extension: MutableMap<Int, Int> = neighborListOf(G, startVertex).associateWithTo(HashMap()) { 1 }
        while (subgraph.vertexCount < k) {
            if (vertexAdditionRule(subgraph, sol)) return Solution()
            val nextVertex: Int = extensionPicker(extension)
            neighborListOf(G, nextVertex).forEach {
                if (it !in subgraph.vertexSet())
                    extension[it] = extension.getOrDefault(it, 0) + 1
            }
            extension.remove(nextVertex)
            subgraph.expandSubgraph(nextVertex)
        }
        return Solution(subgraph, f.eval(subgraph))
    }


    val vertexDegreeMap: Map<Int, Int> = G.vertexSet().associateWith { G.degreeOf(it) }


    fun fullRun(start: Int, vertexPicker: (extension: MutableMap<Int, Int>) -> Int) {

        val heuristicSolution = generateStartingSolution(start, vertexPicker)

        while (true) {
            /**Search and update with neighbourhood until no more improvement can be made.*/
            val oldVal = heuristicSolution.value
            localSearchStep(heuristicSolution)
            if (oldVal == heuristicSolution.value) break
        }

        /**The new solution is better and gets saved.*/
        if (heuristicSolution.value > sol.value) {
            sol.subgraph = heuristicSolution.subgraph.copy()
            sol.value = heuristicSolution.value
        }
    }

    repeat(runs) {
        if (sol.value >= f.globalOptimum()) return@repeat
        else {
            //Chooses vertices randomly proportionally to their degree -> Is good for dense graphs.
            fullRun(takeRandom(vertexDegreeMap)) { takeRandom(it) }

            //Chooses vertices randomly inverse-proportionally to their degree -> Is good for sparse graphs.
            val v = takeRandom(vertexDegreeMap) { x -> Inverse().value(x.toDouble()) }
            fullRun(v) { takeRandom(it) { x -> Inverse().value(x.toDouble()) } }

            //Choose vertices completely random, ignoring their degrees (Laplace-Distribution). Works as OK at all times.
            fullRun(vertexDegreeMap.keys.random()) { map -> map.keys.random() }
        }
    }
    println("Heuristic:".padEnd(Settings.paddingRight) + "${"%.1f".format(GlobalData.secondsElapsed())} sec.")
    return sol
}

/**This function takes a intermediate [solution] of the correct size and tries to improve it by replacing it with a
 * solution that can be found within the 1-swap neighbourhood.
 * If there is no subgraph inside the 1-swap neighbourhood with a better value for [f], then [solution] will just stay
 * the same.
 * This tells you that the current iteration of local-search is finished (or that you need to consider a bigger neighbourhood).
 *
 * The 1-swap neighbourhood is defined as: Let *x* be the current subgraph stored in [solution].
 * Then for each vertex *v* of *x* it is tested if [f] (*x*) can be improved by deleting *v* and instead adding another
 * vertex *w* to the subgraph. *w* then gets an edge to edge vertex in *x* it also has an edge to inside of *g*.
 *
 * This function also ensures that *x* is connected after adding *w*.
 * The value of the new-found subgraph also gets stored in [solution].*/
fun  localSearchStep(solution: Solution<Int>) {
    for (deletedVertex: Int in solution.subgraph.vertexSet().toList()) { //needs to copy to prevent ConcurrentModifierException
        solution.subgraph.removeVertex(deletedVertex)
        for (newVertex in intersectAll(ConnectivityInspector(solution.subgraph).connectedSets().map { G.neighbours(it) })) {
            solution.subgraph.expandSubgraph(newVertex)

            if (f.eval(solution.subgraph) > solution.value) {   //new solution is better -> update and end
                solution.value = f.eval(solution.subgraph)
                return
            }
            solution.subgraph.removeVertex(newVertex)
        }
        solution.subgraph.expandSubgraph( deletedVertex)
    }
}
