package de.umr.fixcon.objectiveFunctions

import de.umr.core.dataStructures.vertexCount
import de.umr.fixcon.AcyclicFunction
import de.umr.fixcon.DegreeConstrainedFunction
import de.umr.fixcon.GlobalData.k
import org.jgrapht.Graph
import org.jgrapht.graph.DefaultEdge


/****FixCon** can optimize for arbitrary target-functions. But for each function, you need to know certain properties of
 * it (for example the maximal value it can attain on a graph of size [k]).
 * As sample of functions is already implemented, together with said properties.
 *
 * What function should be used is specified by an ID via the console. Here the target-function is returned according
 * to the ID.*/
fun graphFunctionByID(id: Int, parameters: List<Int> = emptyList()) =
        when (id) {
            1 -> EdgeCountFunction()
            2 -> MinDegreeFunction()
            3 -> NegMaxDegreeFunction()
            4 -> AcyclicFunction()
            5 -> TriangleFreeFunction()
            6 -> DiameterFunction()
            7 -> RRegularFunction(parameters)
            8 -> DegreeConstrainedFunction(parameters)
            else -> throw IllegalArgumentException("No function exists for this id: $id")
        }

/**Specifies a type any function that maps a finite graph to a real number must fulfill.
 *
 * @param params Most instances of [AbstractGraphFunction] don't depend on any
 * parameters (like number of edges or maximal degree), in this case [params] is empty.
 * But if it does (like [DegreeConstrainedFunction] and [RRegularFunction], they are specified here.
 */
abstract class AbstractGraphFunction(val params: List<Int> = emptyList()) {

    /**Shows if the [AbstractGraphFunction] uses the edge-weights of a graph to compute [eval]
     * This property may also be useful as some pruning-rules only make sense on NOT edge-weighted objective functions.*/
    open val isEdgeWeighted: Boolean = false

    /**True iff in the graph you can replace any vertex with one of its twins and the value of [eval] remains unchanged.
     * (Is usually false if the function is dependent on weights of edges).*/
    open val allowsTwinRule: Boolean = false

    /**Let *g'* be a graph created by adding edges, but no vertices to *g*.
     * Then *f* is edge-monotone if *f*(*g'*) is not lower than *f*(*g*)*/
    open val edgeMonotone: Boolean = false

    /**Returns how much the output-value of this function can increase at most by adding
     * exactly one vertex to the graph.*/
    open val vertexAdditionBound: Int get() = 0     //Holds true for all hereditary functions (they can only decrease).

    /**A target function *f* is vertex-addition-bounded by value *x*, if for every graph *G* and
    all graphs *G'* that are obtained by adding some vertex to *G* and making this vertex adjacent to
    some subset of *Vertices(G)*, we have *f(G')* less-or-equal *f(G)* + *x*.*/
    open fun <V> completeBound(subgraph: Graph<V, DefaultEdge>) = (k - subgraph.vertexCount) * vertexAdditionBound

    /**Applies the function to a graph and returns the resulting real number*/
    abstract fun <V> eval(g: Graph<V, DefaultEdge>): Int

    /**Returns the optimum value the function can return for a graph of the size [k]
     * [k] has a default-value because for some functions it's not needed, e.g. for decision-problems
     * The default value of this function is 1, which is the [globalOptimum] for decision-problems*/
    abstract fun globalOptimum(): Int
}